# -*- coding: utf-8 -*-

from collections import OrderedDict
import os
import sys
try:
    from pysqlite2 import dbapi2 as sqlite3
except:
    try:
        import sqlite3
    except:
        print 'sqlite module needs to be installed'
        sys.exit()

import time

from . import G2GValueError
from .g2g_utils import configure_logging, format_time, get_logger, Location
from .gtf import GTF
import g2g_fileutils as g2g_fu

LOG = get_logger()

SQL_CREATE_GTF_TABLE = """
    CREATE TABLE gtf (
        _key INTEGER PRIMARY KEY,
        gene_id TEXT,
        transcript_id TEXT,
        ensembl_id TEXT,
        seqid TEXT NOT NULL,
        start INTEGER NOT NULL,
        end INTEGER NOT NULL,
        strand INTEGER,
        score TEXT,
        source_key INTEGER NOT NULL,
        type_key INTEGER NOT NULL,
        frame TEXT
    )
"""

SQL_CREATE_GTF_LOOKUP_TABLE = """
    CREATE TABLE gtf_lookup (
        _key INTEGER PRIMARY KEY,
        gtf_key INTEGER NOT NULL,
        attribute_key TEXT NOT NULL,
        value TEXT NOT NULL
    )
"""

SQL_CREATE_GTF_TYPES_TABLE = """
    CREATE TABLE gtf_types (
        _key INTEGER PRIMARY KEY,
        gtf_type TEXT NOT NULL
    )
"""

SQL_CREATE_GTF_SOURCES_TABLE = """
    CREATE TABLE gtf_sources (
        _key INTEGER PRIMARY KEY,
        gtf_source TEXT NOT NULL
    )
"""

SQL_CREATE_GTF_ATTRIBUTES_TABLE = """
    CREATE TABLE gtf_attributes (
        _key INTEGER PRIMARY KEY,
        gtf_attribute TEXT NOT NULL
    )
"""

SQL_INSERT_GTF_TABLE = """
    INSERT
      INTO gtf
    VALUES (null, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
"""

SQL_INSERT_GTF_LOOKUP_TABLE = """
    INSERT
      INTO gtf_lookup
    VALUES (null, ?, ?, ?)
"""

SQL_INSERT_GTF_TYPES_TABLE = """
    INSERT
      INTO gtf_types
    VALUES (?, ?)
"""

SQL_INSERT_GTF_SOURCES_TABLE = """
    INSERT
      INTO gtf_sources
    VALUES (?, ?)
"""

SQL_INSERT_GTF_ATTRIBUTES_TABLE = """
    INSERT
      INTO gtf_attributes
    VALUES (?, ?)
"""

SQL_INDICES_GTF = [
    "CREATE INDEX idx_gtf_gene_id ON gtf(gene_id ASC)",
    "CREATE INDEX idx_gtf_transcript_id ON gtf(transcript_id ASC)",
    "CREATE INDEX idx_gtf_ensembl_id ON gtf(ensembl_id ASC)",
    "CREATE INDEX idx_gtf_seqid ON gtf(seqid ASC)",
    "CREATE INDEX idx_gtf_start ON gtf(start ASC)",
    "CREATE INDEX idx_gtf_end ON gtf(end ASC)",
    "CREATE INDEX idx_gtf_strand ON gtf(strand ASC)",
    "CREATE INDEX idx_gtf_score ON gtf(score ASC)",
    "CREATE INDEX idx_gtf_source_key ON gtf(source_key ASC)",
    "CREATE INDEX idx_gtf_type_key ON gtf(type_key ASC)",
    "CREATE INDEX idx_gtf_frame ON gtf(frame ASC)"
]

SQL_INDICES_GTF_LOOKUP = [
    "CREATE INDEX idx_gtf_gtf_key ON gtf_lookup(gtf_key ASC)",
    "CREATE INDEX idx_gtf_lookup_attribute_key ON gtf_lookup(attribute_key ASC)",
    "CREATE INDEX idx_gtf_lookup_value ON gtf_lookup(value ASC)"
]

SQL_INDICES_GTF_TYPES = [
    "CREATE INDEX idx_gtf_types_gtf_type ON gtf_types(gtf_type ASC)",
]

SQL_INDICES_GTF_SOURCES = [
    "CREATE INDEX idx_gtf_sources_gtf_source ON gtf_sources(gtf_source ASC)",
]

SQL_INDICES_GTF_ATTRIBUTES = [
    "CREATE INDEX idx_gtf_attributes_gtf_attribute ON gtf_attributes(gtf_attribute ASC)",
]


SQL_GENE_FOR_ANY = """
SELECT g._key, g.ensembl_id, g.gene_id, g.transcript_id, g.seqid, s.gtf_source, t.gtf_type, g.seqid, g.start, g.end, g.strand, a.gtf_attribute, l.value
  FROM gtf g, gtf g2, gtf_lookup l, gtf_attributes a, gtf_sources s, gtf_types t
 WHERE g._key = l.gtf_key
   AND g.source_key = s._key
   AND g.type_key = t._key
   AND l.attribute_key = a._key
   AND g.gene_id = g2.gene_id
   and g2.ensembl_id = :ensembl_id
 ORDER BY g._key
"""

SQL_GENES_SIMPLE = """
SELECT distinct g.ensembl_id, g.seqid, g.start, g.end, g.strand
  FROM gtf g
 WHERE g.gene_id is not null
   AND g.transcript_id is null
"""
SQL_GENES_SIMPLE_ORDER_BY = " ORDER BY g._key "

# TODO need to get exons and transcripts

SQL_TRANSCRIPTS_SIMPLE = """
SELECT g._key, g.ensembl_id transcript_id, g.seqid transcript_seqid, g.start transcript_start, g.end transcript_end, g.strand transcript_strand, t.gtf_type transcript_type,
       g2._key _child_key, g2.ensembl_id child_id, g2.seqid child_seqid, g2.start child_start, g2.end child_end, g2.strand child_strand, t2.gtf_type child_type,
       g.gene_id, a.gtf_attribute, l.value
  FROM gtf g, gtf g2, gtf_types t, gtf_types t2, gtf_lookup l, gtf_attributes a
 WHERE g.type_key = t._key
   AND g.transcript_id is not null
   AND t.gtf_type = 'transcript'
   AND g2.transcript_id = g.transcript_id
   AND g2.type_key = t2._key
   AND t2.gtf_type in ('exon', 'transcript')
   AND l.gtf_key = g2._key
   AND l.attribute_key = a._key
   AND a.gtf_attribute in ('exon_number', 'gene_name')
"""

SQL_TRANSCRIPTS_SIMPLE_ORDER_BY = " ORDER BY l._key "

def count_me(conn, table):
    c = conn.cursor()
    c.execute("select count(1) from {}".format(table))
    for row in c:
        print str(row)
    c.close()

def gtf2db(input_file, output_file):
    """
    Convert a GTF file into SQLite

    :param input_file: the GTF file to convert
    :param output_file: The generated database file
    """
    start = time.time()

    input_file = g2g_fu.check_file(input_file, 'r')
    output_file = g2g_fu.check_file(output_file, 'w')

    g2g_fu.delete_file(output_file)

    LOG.info("GTF FILE: {0}".format(input_file))
    LOG.info("DB File: {0}".format(output_file))

    conn = sqlite3.connect(output_file)
    c = conn.cursor()

    LOG.debug("Generating tables")
    c.execute(SQL_CREATE_GTF_TABLE)
    c.execute(SQL_CREATE_GTF_LOOKUP_TABLE)
    c.execute(SQL_CREATE_GTF_SOURCES_TABLE)
    c.execute(SQL_CREATE_GTF_TYPES_TABLE)
    c.execute(SQL_CREATE_GTF_ATTRIBUTES_TABLE)



    gtf_types = {}
    gtf_sources = {}
    gtf_attributes = {}

    LOG.info("Parsing GTF file...")

    gtf_file = GTF(input_file)

    counter = 0

    for record in gtf_file:
        if counter and counter % 100000 == 0:
            LOG.info("Processed {0:,} records".format(counter))

        if record.type not in gtf_types:
            _type_key = len(gtf_types.keys())
            gtf_types[record.type] = _type_key
        else:
            _type_key = gtf_types[record.type]

        if record.source not in gtf_sources:
            _source_key = len(gtf_sources.keys())
            gtf_sources[record.source] = _source_key
        else:
            _source_key = gtf_sources[record.source]

        strand = 0
        if record.strand in ['+', '-']:
            strand = 1 if record.strand == '+' else -1

        gene_id = record.attributes['gene_id']
        transcript_id = record.attributes['transcript_id'] if 'transcript_id' in record.attributes else None
        ensembl_id = None

        if record.type == 'gene':
            ensembl_id = record.attributes['gene_id']
        elif record.type == 'transcript':
            ensembl_id = record.attributes['transcript_id']
        elif record.type == 'exon':
            ensembl_id = record.attributes['exon_id']
        else:
            ensembl_id = record.attributes['protein_id'] if 'protein_id' in record.attributes else None

        c.execute(SQL_INSERT_GTF_TABLE, (gene_id, transcript_id, ensembl_id, record.seqid, record.start, record.end, strand, record.score, _source_key, _type_key, record.frame))
        gtf_key = c.lastrowid

        for attribute, value in record.attributes.iteritems():
            if attribute not in ['gene_id', 'transcript_id', 'exon_id']:
                if attribute not in gtf_attributes:
                    _attribute_key = len(gtf_attributes.keys())
                    gtf_attributes[attribute] = _attribute_key
                else:
                    _attribute_key = gtf_attributes[attribute]

                c.execute(SQL_INSERT_GTF_LOOKUP_TABLE, (gtf_key, _attribute_key, value))

        counter += 1

    # save (commit) the changes
    conn.commit()

    for source, _key in gtf_sources.iteritems():
        c.execute(SQL_INSERT_GTF_SOURCES_TABLE, (_key, source))
        conn.commit()

    for type, _key in gtf_types.iteritems():
        c.execute(SQL_INSERT_GTF_TYPES_TABLE, (_key, type))
        conn.commit()

    for attribute, _key in gtf_attributes.iteritems():
        c.execute(SQL_INSERT_GTF_ATTRIBUTES_TABLE, (_key, attribute))
        conn.commit()

    LOG.info("GTF File parsed")

    LOG.info("Finalizing database...")

    for sql in SQL_INDICES_GTF:
        LOG.debug(sql)
        c.execute(sql)

    for sql in SQL_INDICES_GTF_LOOKUP:
        LOG.debug(sql)
        c.execute(sql)

    for sql in SQL_INDICES_GTF_TYPES:
        LOG.debug(sql)
        c.execute(sql)

    for sql in SQL_INDICES_GTF_SOURCES:
        LOG.debug(sql)
        c.execute(sql)

    for sql in SQL_INDICES_GTF_ATTRIBUTES:
        LOG.debug(sql)
        c.execute(sql)

    LOG.info("Database created")

    # close connection
    conn.close()

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


class GTFObject(object):
    def __init__(self, ensembl_id=None, seqid=None, start=None, end=None, strand=None):
        self.ensembl_id = ensembl_id
        self.seqid = seqid
        self.start = start
        self.end = end
        self.strand = strand
        self.source = None
        self.name = None
        self.biotype = None


class Gene(GTFObject):
    def __init__(self, ensembl_id=None, seqid=None, start=None, end=None, strand=None):
        GTFObject.__init__(self, ensembl_id, seqid, start, end, strand)
        self.transcripts = OrderedDict()

    def __str__(self):
        return "Gene: {0} {1}:{2}-{3} ({4})".format(self.ensembl_id, self.seqid, self.start, self.end, self.strand)


class Transcript(GTFObject):
    def __init__(self, ensembl_id=None, seqid=None, start=None, end=None, strand=None):
        GTFObject.__init__(self, ensembl_id, seqid, start, end, strand)
        self.gene_ids = OrderedDict()
        self.exons = OrderedDict()

    def __str__(self):
        return "Transcript: {0} {1}:{2}-{3} ({4})".format(self.ensembl_id, self.seqid, self.start, self.end, self.strand)


class Exon(GTFObject):
    def __init__(self, ensembl_id=None, seqid=None, start=None, end=None, strand=None):
        GTFObject.__init__(self, ensembl_id, seqid, start, end, strand)
        self.gene_id = None
        self.transcript_ids = OrderedDict()
        self._exon_number = None

    @property
    def exon_number(self):
        return self._exon_number

    @exon_number.setter
    def exon_number(self, value):
        if value:
            try:
                self._exon_number = int(value)
            except ValueError, ve:
                raise G2GValueError("Illegal value for exon_number {0}, start must be an integer".format(value))

    def __str__(self):
        en = ''
        if self.exon_number:
            en = " #{0}".format(self.exon_number)
        return "Exon: {0} {1}:{2}-{3} ({4}){5}".format(self.ensembl_id, self.seqid, self.start, self.end, self.strand, en)



def location_to_sql(location, use_strand=False, overlap=True):
    sql = ""
    sql_parameters = {}

    if location:
        sql = " AND g.seqid = :seqid "
        sql_parameters = {'seqid': location.seqid}

        if overlap:
            if location.start and not location.end:
                sql = "{0} AND g.start <= :end AND g.end >= :start ".format(sql)
                sql_parameters['start'] = location.start
                sql_parameters['end'] = location.start
            elif not location.start and location.end:
                sql = "{0} AND g.start <= :end AND g.end >= :start ".format(sql)
                sql_parameters['start'] = location.end
                sql_parameters['end'] = location.end
            elif location.start and location.end:
                sql = "{0} AND g.start <= :end AND g.end >= :start ".format(sql)
                sql_parameters['start'] = location.start
                sql_parameters['end'] = location.end
        else:
            if location.start:
                sql = "{0} AND g.start >= :start ".format(sql)
                sql_parameters['start'] = location.start

            if location.end:
                sql = "{0} AND g.end <= :end ".format(sql)
                sql_parameters['end'] = location.end

        if use_strand:
            if location.strand == '-':
                strand = -1
            else:
                strand = 1

            sql = "{0} AND g.strand = :strand ".format(sql)
            sql_parameters['strand'] = strand

    return sql, sql_parameters


def get_gene(db, ensembl_id):
    """
    Retrieve a Gene object by the ensembl id.

    The ensembl id can be any value (Gene, Transcript, or Exon ID)

    """
    conn = sqlite3.connect(db)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()
    cursor.execute(SQL_GENE_FOR_ANY, {'ensembl_id': ensembl_id})

    genes = OrderedDict()
    transcripts = OrderedDict()
    exons = OrderedDict()

    for r in cursor:
        _key = r['_key']
        gtf_type = r['gtf_type']
        attribute = r['gtf_attribute']
        value = r['value']

        if gtf_type == 'gene':
            gene = genes.get(_key, Gene(r['ensembl_id'], r['seqid'], r['start'], r['end'], r['strand']))

            if attribute == 'gene_name':
                gene.name = value
            elif attribute == 'gene_source':
                gene.source = value

            genes[_key] = gene

        elif gtf_type == 'transcript':
            transcript = transcripts.get(_key, Transcript(r['ensembl_id'], r['seqid'], r['start'], r['end'], r['strand']))
            transcript.gene_ids[r['gene_id']] = r['gene_id']

            if attribute == 'transcript_name':
                transcript.name = value
            elif attribute == 'transcript_source':
                transcript.source = value

            transcripts[_key] = transcript

        elif gtf_type == 'exon':
            exon = exons.get(_key, Exon(r['ensembl_id'], r['seqid'], r['start'], r['end'], r['strand']))
            exon.gene_ids = r['gene_id']
            exon.transcript_ids[r['transcript_id']] = r['transcript_id']

            if attribute == 'exon_number':
                exon.exon_number = value

            exons[_key] = exon

    genes = {gene.ensembl_id: gene for i, gene in genes.iteritems()}
    transcripts = {transcript.ensembl_id: transcript for i, transcript in transcripts.iteritems()}

    for _id, exon in exons.iteritems():
        for _tid in exon.transcript_ids:
            transcripts[_tid].exons[exon.ensembl_id] = exon

    for _id, transcript in transcripts.iteritems():
        for _gid in transcript.gene_ids:
            genes[_gid].transcripts[_id] = transcript

    cursor.close()
    conn.close()
    return genes


def get_genes_ids(db, location=None, use_strand=False, overlap=True):
    """
    Get ensembl ids for genes
    """

    sql, sql_parameters = location_to_sql(location, use_strand, overlap)
    sql = "{0} {1} {2}".format(SQL_GENES_SIMPLE, sql, SQL_GENES_SIMPLE_ORDER_BY)

    LOG.debug("SQL:\n{0}".format(sql))
    LOG.debug("PARAMETERS: {0}".format(sql_parameters))

    conn = sqlite3.connect(db)
    sqlite3.enable_callback_tracebacks(True)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()

    cursor.execute(sql, sql_parameters)

    gene_ids = []
    for r in cursor:
        gene_ids.append(r['ensembl_id'])

    cursor.close()
    conn.close()
    return gene_ids



def get_genes(db, location=None, use_strand=False, overlap=True):
    """
    Get Gene objects
    """
    gene_ids = get_genes_ids(db, location, use_strand, overlap)

    genes = {}
    for i, gene_id in enumerate(gene_ids):
        genes_temp = get_gene(db, gene_id)
        if genes_temp:
            for ensembl_id, gene in genes_temp.iteritems():
                genes[ensembl_id] = gene

    return genes


def get_genes_simple(db, location=None, use_strand=False, overlap=True):

    sql, sql_parameters = location_to_sql(location, use_strand, overlap)
    sql = "{0} {1} {2}".format(SQL_GENES_SIMPLE, sql, SQL_GENES_SIMPLE_ORDER_BY)

    LOG.debug("SQL:\n{0}".format(sql))
    LOG.debug("PARAMETERS: {0}".format(sql_parameters))

    conn = sqlite3.connect(db)
    sqlite3.enable_callback_tracebacks(True)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()

    cursor.execute(sql, sql_parameters)

    genes = []
    for r in cursor:
        genes.append(Gene(r['ensembl_id'], r['seqid'], r['start'], r['end'], r['strand']))

    cursor.close()
    conn.close()

    return genes

def get_transcripts_simple(db, location=None, use_strand=False, overlap=True):
    sql, sql_parameters = location_to_sql(location, use_strand, overlap)
    sql = "{0} {1} {2}".format(SQL_TRANSCRIPTS_SIMPLE, sql, SQL_TRANSCRIPTS_SIMPLE_ORDER_BY)

    LOG.debug("SQL:\n{0}".format(sql))
    LOG.debug("PARAMETERS: {0}".format(sql_parameters))

    conn = sqlite3.connect(db)
    sqlite3.enable_callback_tracebacks(True)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()

    cursor.execute(sql, sql_parameters)

    transcripts = OrderedDict()
    exons = OrderedDict()
    for r in cursor:

        if r['transcript_id'] == r['child_id']:
            # transcript
            if r['_key'] not in transcripts:
                transcripts[r['_key']] = Transcript(r['transcript_id'], r['transcript_seqid'], r['transcript_start'], r['transcript_end'], r['transcript_strand'])
        else:
            # exon
            exon = exons.get(r['_child_key'], Exon(r['child_id'], r['child_seqid'], r['child_start'], r['child_end'], r['child_strand']))
            exon.gene_id = r['transcript_id']
            exon.transcript_ids[r['transcript_id']] = r['transcript_id']
            attribute = r['gtf_attribute']
            value = r['value']

            if attribute == 'exon_number':
                exon.exon_number = value

            exons[r['_child_key']] = exon

    transcripts = {transcript.ensembl_id: transcript for i, transcript in transcripts.iteritems()}

    for _id, exon in exons.iteritems():
        for _tid in exon.transcript_ids:
            transcripts[_tid].exons[exon.ensembl_id] = exon

    cursor.close()
    conn.close()

    return transcripts.values()


if __name__ == '__main__':
    import sys
    configure_logging(5)

    genes = get_gene(sys.argv[1], sys.argv[2])
    for k,v in genes.iteritems():
        print v
        for k1, v1 in v.transcripts.iteritems():
            print "\t", v1
            for k2,v2 in v1.exons.iteritems():
                print "\t\t", v2

    print '*'*80

    genes = get_genes(sys.argv[1], Location('1', 21250000, 21500000, '-'), overlap=True, use_strand=True)
    for k,v in genes.iteritems():
        print v
        for k1, v1 in v.transcripts.iteritems():
            print "\t", v1
            for k2,v2 in v1.exons.iteritems():
                print "\t\t", v2

    print len(genes)

    """
    gene_ids = get_genes_ids(sys.argv[1], Location('1', 21250000, 21500000), overlap=True)
    print len(gene_ids)

    genes = get_genes_simple(sys.argv[1], Location('1', 21250000, 21500000, '-'), overlap=False, use_strand=True)

    for i, gene in enumerate(genes):
        print gene

    for i, transcript in enumerate(get_transcripts_simple(sys.argv[1], Location('12', 4833175, 4866873), overlap=False)):
            print transcript, len(transcript.exons)
            for k2,v2 in transcript.exons.iteritems():
                print "\t", v2

    """


# -*- coding: utf-8 -*-

#
# Collection of g2g utility functions and classes
#

import bz2
import gzip
import os
import time
import urllib
from subprocess import Popen

from pysam import tabix_compress, faidx, tabix_index

from . import G2GValueError

def delete_file(file_name):
    """

    :param file_name:
    :return:
    """
    try:
        os.remove(file_name)
    except:
        pass


def delete_index_files(filename):
    delete_file("{0}.gzi".format(filename))
    delete_file("{0}.fai".format(filename))
    delete_file("{0}.tbi".format(filename))


def bgzip_decompress(filename):
    # gobble the output
    with open(os.devnull, "w") as fnull:
        p = Popen(['bgzip', '-f', '-d', filename], stdout=fnull, stderr=fnull)
        p.wait()

    delete_index_files(filename)


def bgzip_index(original_file, new_file, file_format):
    """

    :param original_file:
    :param new_file:
    :param file_format:
    :return:
    """

    if file_format.lower() == 'fa':
        tabix_compress(original_file, new_file)
        faidx(new_file)
        delete_file(original_file)
    elif file_format.lower() == 'vcf':
        tabix_index(original_file, preset="vcf", force=True)
    else:
        raise G2GValueError("Unknown file format: {0}".format(file_format))


def open_resource(resource, mode='rb'):
    """
    Open different types of files and return the handle.

    :param resource: a file located locally or on the internet.  Gzip'd and zip'd files are handled.
    :param mode: standard file open modes
    :return: the resource (file) handle
    """
    if not resource:
        return None

    if not isinstance(resource, basestring):
        return resource

    if resource.endswith(('.gz', '.Z', '.z')):
        return gzip.open(resource, mode)
    elif resource.endswith(('.bz', '.bz2', '.bzip2')):
        return bz2.BZ2File(resource, mode)
    elif resource.startswith(('http://', 'https://', 'ftp://')):
        return urllib.urlopen(resource)
    else:
        return open(resource, mode)


def gen_file_name(name, output_dir='.', extension='log', append_time=True):
    """
    Generate a file name.

    :param name: a base for the file name
    :param output_dir: the directory
    :param extension: give the fils le an extension
    :param append_time: append time to the file name (before the extension)
    :return: the absolute path to the file
    """
    if extension[-1] == '.':
        extension = extension[-1:]

    if append_time:
        t = time.strftime("%Y-%m-%d.%H_%M_%S")
        file_name = "{0}.{1}.{2}".format(name, t, extension)
    else:
        file_name = "{0}.{1}".format(name, extension)

    return os.path.abspath(os.path.join(output_dir, file_name))



def get_extension(filename):
    file_first, file_extension = os.path.splitext(filename)
    return file_extension


def prepend_before_extension(filename, text):
    file_first, file_extension = os.path.splitext(filename)

    if file_extension:
        return "{0}.{1}{2}".format(file_first, text, file_extension)

    return filename


def check_file(filename, mode='r'):

    if mode == 'r':

        if filename and os.path.exists(filename):
            return os.path.abspath(filename)

        raise G2GValueError("The following file does not exist: {0}".format(filename))

    elif mode == 'w':

        file_dir = '.'

        if filename:
            file_name = os.path.abspath(filename)
            file_dir = os.path.dirname(file_name)

            if not os.access(file_dir, os.W_OK | os.X_OK):
                raise G2GValueError("Cannot generate file: {0}".format(filename))

            return file_name


    raise G2GValueError("Unspecified mode")






if __name__ == '__main__':
    print ''



# -*- coding: utf-8 -*-

#
# Collection of functions related to GTF
#


from collections import namedtuple, OrderedDict
import os
import sys

from .g2g_utils import get_logger
from .exceptions import G2GGTFError, G2GValueError
from .chain import ChainFile

import g2g_fileutils as g2g_fu

gtfInfoFields = ['seqid', 'source', 'type', 'start', 'end', 'score', 'strand', 'frame', 'attributes']
GTFRecord = namedtuple('GTFRecord', gtfInfoFields)

LOG = get_logger()

class GTF(object):
    """
    Simple GTF object for parsing GTF files

    http://blog.nextgenetics.net/?e=27

    Supports transparent gzip decompression.
    """
    def __init__(self, file_name):
        if not file_name:
            raise G2GGTFError("A filename must be supplied")

        self.file_name = file_name
        self.current_line = None
        self.current_record = None
        self.reader = g2g_fu.open_resource(file_name)

    def __iter__(self):
        return self

    def next(self):
        self.current_line = self.reader.next()

        while self.current_line.startswith('#') or self.current_line.startswith('!'):
            self.current_line = self.reader.next()

        self.current_record = parse_gtf_line(self.current_line)

        return self.current_record


def parse_attributes(attributes):
    """
    Parse the GTF attribute column and return a dict
    """
    if attributes == ".":
        return OrderedDict()

    ret = OrderedDict()
    for attribute in attributes.strip().split(";"):
        if len(attribute):
            elems = attribute.strip().split(' ')
            key = elems[0]
            val = ' '.join(elems[1:])
            if val[0] == '"':
                val = val[1:]
            if val[-1] == '"':
                val = val[0:-1]
            ret[key] = val

    return ret


def parse_gtf_line(line):
    """
    Parse the GTF line.

    :param line: a line from GTF file
    :type line: str
    :return:
    """
    elem = line.strip().split("\t")

    # If this fails, the file format is not standard-compatible
    if len(elem) != len(gtfInfoFields):
        LOG.error(line)
        LOG.error("{0} != {1}".format(len(elem), len(gtfInfoFields)))
        raise G2GGTFError("Improperly formatted GTF file")

    data = {
        'seqid': None if elem[0] == '.' else elem[0].replace('"', ''),
        'source': None if elem[1] == '.' else elem[1].replace('"', ''),
        'type': None if elem[2] == '.' else elem[2].replace('"', ''),
        'start': None if elem[3] == '.' else int(elem[3]),
        'end': None if elem[4] == '.' else int(elem[4]),
        'score': None if elem[5] == '.' else float(elem[5]),
        'strand': None if elem[6] == '.' else elem[6].replace('"', ''),
        'frame': None if elem[7] == '.' else elem[7].replace('"', ''),
        'attributes': parse_attributes(elem[8])
    }

    return GTFRecord(**data)


def convert_gtf_file(chain_file, input_file, output_file, reverse=False):
    """
    Convert GTF coordinates.

    The mappings of coordinates are stored in the :class:`.chain.ChainFile` object.

    :param chain_file:
    :type chain_file: :class:`.chain.ChainFile`
    :param input_file: the input GTF file
    :type input_file: string
    :param output_file: the output GTF file
    :type output_file: string
    :param reverse: reverse direction of original chain file
    :type reverse: boolean
    :return:
    """
    if not isinstance(chain_file, ChainFile):
        chain_file = g2g_fu.check_file(chain_file)

    input_file = g2g_fu.check_file(input_file)
    output_file_name = g2g_fu.check_file(output_file, 'w')
    unmapped_file_name = "{0}.unmapped".format(output_file_name)

    LOG.info("CHAIN FILE: {0}".format(chain_file))
    LOG.info("INPUT FILE: {0}".format(input_file))
    LOG.info("OUTPUT FILE: {0}".format(output_file_name))
    LOG.info("UNMAPPED FILE: {0}".format(unmapped_file_name))

    if not isinstance(chain_file, ChainFile):
        LOG.info("Parsing chain file...")
        chain_file = ChainFile(chain_file, reverse=reverse)
        LOG.info("Chain file parsed")

    gtf_out = open(output_file_name, "w")
    gtf_unmapped_file = open(unmapped_file_name, "w")

    LOG.info("Converting GTF file...")

    gtf_file = GTF(input_file)

    total = 0
    success = 0
    fail = 0

    # GTF is 1 based, bx-python is 0 based
    # when we do the querying, we subtract 1 from the GTF file start position
    # K.B.  Also note in gtf when (s, e) is given...it should mean s <= x <= e.
    #       bx-python (s, e) does it s <= x < e.

    for record in gtf_file:

        LOG.debug("\nORIGINAL: {0}".format(str(gtf_file.current_line).strip()))

        total += 1

        if total % 100000 == 0:
            LOG.info("Processed {0:,} lines".format(total))

        mappings = chain_file.find_mappings(record.seqid, record.start - 1, record.end)

        # unmapped
        if mappings is None:
            LOG.debug("\tFail due to no mappings")
            gtf_unmapped_file.write(gtf_file.current_line)
            fail += 0
            continue
        else:
            LOG.debug("{0} mappings found".format(len(mappings)))

        success += 1
        start = mappings[0].to_start + 1
        end = mappings[-1].to_end

        LOG.debug("({0}, {1}) => ({2}, {3})".format(record.start - 1, record.end, start, end))

        elems = gtf_file.current_line.rstrip().split('\t')
        elems[3] = start
        elems[4] = end

        LOG.debug("     NEW: {0}".format("\t".join(map(str, elems))))

        gtf_out.write("\t".join(map(str, elems)))
        gtf_out.write("\n")

    gtf_out.close()
    gtf_unmapped_file.close()

    LOG.info("Converted {0:,} of {1:,} records".format(success, total))
    LOG.info('GTF file converted')




#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from distutils.extension import Extension

from setuptools import *
from glob import glob


try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True

from distutils.core import Command

# Use build_ext from Cython
command_classes = {}

# Use build_ext from Cython if found
try:
    import Cython.Distutils
    command_classes['build_ext'] = Cython.Distutils.build_ext
except:
    pass

# Run 2to3 builder if we're on Python 3.x, from
#   http://wiki.python.org/moin/PortingPythonToPy3k
try:
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:
    # 2.x
    from distutils.command.build_py import build_py
command_classes['build_py'] = build_py

requirements = []

with open('requirements.txt') as requirements_file:
    for line in requirements_file:
        requirements.append(line.strip())

test_requirements = [
    # TODO: put package test requirements here
]


import os
on_rtd = os.environ.get('READTHEDOCS', None)

if not on_rtd:
    requirements.append('bx-python>=0.7.2')
    requirements.append('pysam>=0.8.1')
    requirements.append('biopython>=1.63')
    requirements.append('pysqlite>=2.6.3')

import g2gtools


def main():
    with open('README.rst') as readme_file:
        readme = readme_file.read()

    with open('HISTORY.rst') as history_file:
        history = history_file.read().replace('.. :changelog:', '')

    setup(
        name='g2gtools',
        version=g2gtools.__version__,
        description="A set of tools that facilitates genome to genome conversion for studying multiparent populations",
        long_description=readme + '\n\n' + history,
        author='Kwangbom "KB" Choi & Matthew Vincent, The Jackson Laboratory',
        author_email='kb.choi@jax.org',
        url='https://github.com/jax-cgd/g2gtools',
        packages = find_packages('.'),
        package_dir={'g2gtools': 'g2gtools'},
        scripts=glob("bin/*"),
        ext_modules=get_extension_modules(),
        cmdclass=command_classes,
        include_package_data=True,
        install_requires=requirements,
        license="GPLv3",
        keywords='g2gtools',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7'
        ],
        test_suite='tests',
        tests_require=test_requirements,
        zip_safe=False
    )


def get_extension_modules():
    extensions = []

    # fasta reading speedups
    #extensions.append(Extension("g2gtools._fasta", ["lib/g2gtools/_fasta.pyx", "lib/g2gtools/Fasta.cpp"], language='c++'))

    return extensions

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#
# Collection of module errors
#


class G2GError(Exception):
    """
    Simple base exception, root of all G2G exceptions
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GError, self).__init__(self.msg)


class G2GValueError(G2GError):
    """
    Generic value error for G2G
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GValueError, self).__init__(self.msg)


class G2GBAMError(G2GError):
    """
    BAM/SAM file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GBAMError, self).__init__(self.msg)


class G2GBedError(G2GError):
    """
    BED file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GBedError, self).__init__(self.msg)


class G2GChainFileError(G2GError):
    """
    Chain file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GChainFileError, self).__init__(self.msg)


class G2GCigarFormatError(G2GError):
    """
    Cigar exception/error
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GCigarFormatError, self).__init__(self.msg)


class G2GFastaError(G2GError):
    """
    Fasta file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GFastaError, self).__init__(self.msg)


class G2GFetchError(G2GError):
    """
    G2G Fetch errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GFetchError, self).__init__(self.msg)


class G2GLocationError(G2GError):
    """
    Location errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GLocationError, self).__init__(self.msg)


class G2GVCFError(G2GError):
    """
    VCF file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GVCFError, self).__init__(self.msg)


class G2GGTFError(G2GError):
    """
    GTF file errors
    """
    def __init__(self, msg=None):
        self.msg = msg
        super(G2GGTFError, self).__init__(self.msg)






# -*- coding: utf-8 -*-

#
# A way to patch a Fasta file quickly
#

import mmap
import multiprocessing
import os
import shutil
import time

import pysam

from .g2g_utils import format_time, get_logger
import g2g_fileutils as g2g_fu
from . import G2GValueError, G2GVCFError, G2GError
from .vcf import VCF, parse_gt, parse_vcf_line
from .fasta import FAI

LOG = get_logger()

NUM_CHUNKS = 10

# globals, really... yup :-(
fasta_index = None
mm_l = None
mm_r = None




class KeyboardInterruptError(Exception): pass






def get_pos(fai, chromosome, start, end):
    """
    Get the byte positions of the fasta file using the specified location

    :param fasta_index: fasta index file used for sequence retrieval
    :type fasta_index: :class:`.fasta.FAI`
    :param chromosome: chromosome to use
    :type chromosome: string
    :param start: starting position
    :type start: int
    :param end: ending position
    :type end: int
    :return: a tuple of starting byte, ending byte, and byte length
    """
    chrom = fai.records[chromosome]
    fai_entry_length = chrom.length
    fai_entry_offset = chrom.offset
    fai_entry_line_length = chrom.line_length
    fai_entry_line_length_bytes = chrom.line_length_bytes
    seq_len = end - start
    line_ratio = fai_entry_line_length * (fai_entry_line_length_bytes - fai_entry_line_length)
    newlines_total = int(fai_entry_length / line_ratio)
    newlines_before = 0
    if start > 0:
        newlines_before = int(start / line_ratio)
    newlines_to_end = int(end / line_ratio)
    byte_len_seq = newlines_to_end - newlines_before + seq_len
    byte_start = fai_entry_offset + newlines_before + start
    byte_end = fai_entry_offset + newlines_total + fai_entry_length
    return byte_start, byte_end, byte_len_seq


def process_piece(filename_vcf, chrom, start, end, sample_index, diploid, pass_only, quality):
    """
    Process a section of the VCF file specified by the location.

    :param filename_vcf: name of the VCF file
    :type filename_vcf: name of the VCF file
    :param chrom: the chromosome
    :type chrom: string
    :param start: starting position
    :type start: int
    :param end: ending position
    :type end: int
    :param sample_index: which sample index to process
    :type sample_index: int
    :param diploid: `True` to process diploid
    :type diploid: boolean
    :param pass_only: `True` to only process 'PASS' VCF records
    :type pass_only: boolean
    :param quality:
    :type quality: boolean
    :return: a dict of 2 elements, 'chrom' and 'count' specifying how many SNPs were patched
    """
    ret = {'count': 0, 'chrom': chrom}

    try:
        query = "{0}:{1}-{2}".format(chrom, start, end)
        LOG.info("Processing {0}...".format(query))
        tb = pysam.TabixFile(filename_vcf)

        global fasta_index
        for d in tb.fetch(chrom, start, end):
            rec = parse_vcf_line(d)

            if pass_only and 'PASS' not in rec.FILTER:
                continue

            byte_start, byte_end, byte_len_seq = get_pos(fasta_index, rec.CHROM, rec.POS-1, rec.POS)
            gt = parse_gt(rec, sample_index)

            #LOG.debug(rec)
            #LOG.debug(gt)

            # FI : Whether a sample was a Pass(1) or fail (0) based on FILTER values
            if quality and gt.fi == '0':
                continue

            global mm_l
            global mm_r
            if gt.left and gt.right:
                if mm_l[byte_start] != gt.ref:
                    LOG.warn("Reference at {0}:{1} has base '{2}', but VCF indicates '{3}', skipping".format(rec.CHROM, rec.POS, mm_l[byte_start], gt.ref))
                    continue

                if diploid:
                    mm_l[byte_start] = gt.left
                    mm_r[byte_start] = gt.right
                    ret['count'] += 1
                    # LOG.debug("Patching {0}:{1} ({2}) from {3} to {4},{5}".format(chrom, rec.POS-1, from_base, gt.ref, gt.left, gt.right))
                else:
                    if gt.left == gt.right:
                        # ignore heterozygotes 0/1, 1/0, only process 0/0 and 1/1
                        mm_l[byte_start] = gt.left
                        ret['count'] += 1
                        #LOG.debug("Patching {0}:{1} from {2} to {3}:{4}".format(chrom, rec.POS-1, gt.ref, gt.left, gt))


    except KeyboardInterrupt:
        raise KeyboardInterruptError()
    except ValueError, te:
        LOG.debug("No SNPS found in region {0}".format(query))
    except Exception, e:
        LOG.error('-'*80)
        LOG.error(e)
        LOG.debug("{0}, {1}, {2}".format(byte_start, byte_end, byte_len_seq))
        LOG.debug(rec)
        LOG.debug(gt)

    return ret


def wrapper(args):
    """
    Simple wrapper, useful for debugging.

    :param args: the arguments to process_piece
    :return: the same as process_piece
    """
    return process_piece(*args)


def patch(filename_original_fasta, filename_vcf, strain, filename_new_fasta_l, filename_new_fasta_r,
          num_processes, pass_only, quality, diploid):
    """
    Perform patch via multiprocess.

    :param filename_vcf: name of the VCF file
    :type filename_vcf: string
    :param strain: name of strain to use in VCF file
    :type strain: string
    :param filename_original_fasta: name of the input Fasta file
    :type filename_original_fasta: string
    :param filename_new_fasta_l: name of the output Fasta file
    :type filename_new_fasta_l: string
    :param filename_new_fasta_r: name of the output Fasta file
    :type filename_new_fasta_r: string
    :param num_processes: the number of processes to spawn
    :type num_processes: int
    :param pass_only: Only process those VCF records with a 'PASS'
    :type pass_only: boolean
    :param quality: filter on quality, FI=PASS
    :type quality: boolean
    :param diploid: don't ignore hets and create 2 files
    :type diploid: boolean
    :return: Nothing
    """

    vcf_file = VCF(filename_vcf)

    # the temporary fasta index
    global fasta_index
    fasta_index = FAI(filename_new_fasta_l)

    if strain not in vcf_file.samples:
        raise G2GVCFError("Unknown strain '{0}', valid strains are: {1}".format(strain, vcf_file.samples.keys()))

    sample_index = vcf_file.get_sample_index(strain)

    fd_l = open(filename_new_fasta_l, "r+b")
    global mm_l
    mm_l = mmap.mmap(fd_l.fileno(), 0)

    if diploid:
        fd_r = open(filename_new_fasta_r, "r+b")
        global mm_r
        mm_r = mmap.mmap(fd_r.fileno(), 0)

    all_start_pos = []
    all_end_pos = []
    all_chrom = []

    try:
        for f in fasta_index.records:
            chrom = f
            chrom_length = fasta_index.records[f].length

            # we don't want to handle to much data at a time, so chunk
            initial_chunks = range(1, chrom_length, chrom_length/NUM_CHUNKS)

            start_pos = sorted(set([i for i in initial_chunks]))
            end_pos = [i-1 for i in start_pos][1:] + [chrom_length]
            chroms = [chrom] * len(start_pos)

            all_start_pos.extend(start_pos)
            all_end_pos.extend(end_pos)
            all_chrom.extend(chroms)

        all_sample_index = [sample_index] * len(all_start_pos)
        all_diploids = [diploid] * len(all_start_pos)
        all_vcffiles = [filename_vcf] * len(all_start_pos)
        all_passonly = [pass_only] * len(all_start_pos)
        all_quality = [quality] * len(all_start_pos)

        args = zip(all_vcffiles, all_chrom, all_start_pos, all_end_pos,
                   all_sample_index, all_diploids, all_passonly, all_quality)

        pool = multiprocessing.Pool(num_processes)
        results = pool.map(wrapper, args)

        # parse results
        total = 0
        chroms = {}

        mm_l.flush()
        mm_l.close()

        if diploid:
            mm_r.flush()
            mm_r.close()

        for c in results:
            current = chroms.get(c['chrom'], 0)
            chroms[c['chrom']] = current + c['count']
            total += c['count']

        for f in fasta_index.records:
            LOG.info("Patched {0:,} SNPs on chromosome {1}".format(chroms.get(f, 0), f))

        LOG.info("Patched {0:,} SNPs total".format(total))
    except KeyboardInterrupt:
        pool.terminate()
        raise G2GError("Execution halted")
    except Exception, e:
        raise G2GError("Execution halted")


def prepare_fasta_patch(filename_fasta, filename_output, bgzip=False, diploid=False):
    """
    Initialize fasta_patch variables

    :param filename_fasta:
    :param filename_vcf:
    :param strain:
    :param filename_output:
    :param bgzip:
    :param diploid:
    :return:
    """

    filename_output = g2g_fu.check_file(filename_output, 'w')
    output_file_dir = os.path.abspath(os.path.dirname(filename_output))

    new_filename_output = filename_output

    # let's figure out what our output names will be
    if filename_output.lower().endswith('.gz'):
        # strip off .gz
        new_filename_output = filename_output[:-3]

    if not filename_output.lower().endswith('.fa'):
        raise G2GValueError("Expecting output filename extension to be either '.fa.gz' or '.fa'")


    if diploid:
        filename_output_l = g2g_fu.prepend_before_extension(new_filename_output, 'l')
        filename_output_r = g2g_fu.prepend_before_extension(new_filename_output, 'r')

        g2g_fu.delete_index_files(filename_output_l)
        g2g_fu.delete_index_files(filename_output_r)
    else:
        filename_output_l = new_filename_output
        filename_output_r = None

        g2g_fu.delete_index_files(filename_output_l)

    # at this point we are hoping for a .fa extension

    # let's figure out our input and process accordingly
    if filename_fasta.lower().endswith('.fa.gz'):
        # decompress the fasta file if it is compressed

        LOG.info("Copying and decompressing fasta file")

        # copy file and preserve gz extension for bgzip -d to work
        tmp_file_name = os.path.basename(filename_fasta)                        # something.gz
        LOG.debug("tmp_file_name={0}".format(tmp_file_name))

        tmp_fasta = os.path.join(output_file_dir, tmp_file_name)                # /path/something.fa.gz
        LOG.debug("tmp_fasta={0}".format(tmp_fasta))

        LOG.debug("COPYING {0} to {1}".format(filename_fasta, tmp_fasta))
        shutil.copy(filename_fasta, tmp_fasta)  # cp /original/something.fa.gz /output/something.fa.gz

        LOG.debug("DECOMPRESSING {0}".format(tmp_fasta))
        g2g_fu.bgzip_decompress(tmp_fasta)

        tmp_fasta = tmp_fasta[:-3]         # /path/something.fa
        LOG.debug("tmp_fasta={0}".format(tmp_fasta))

        LOG.debug("Moving '{0}' to '{1}'...".format(tmp_fasta, filename_output_l))
        shutil.move(tmp_fasta, filename_output_l)

    elif filename_fasta.lower().endswith('.fa'):
        LOG.debug("File is not compressed")

        LOG.debug("COPYING {0} to {1}".format(filename_fasta, filename_output_l))
        shutil.copy(filename_fasta, filename_output_l)
    else:
        raise G2GValueError("Expecting input filename extension to be either '.fa.gz' or '.fa'")

    if diploid:
        LOG.debug("Copying '{0}' to '{1}'...".format(filename_output_l, filename_output_r))
        shutil.copy(filename_output_l, filename_output_r)

    # build a temporary fasta index
    pysam.FastaFile(filename_output_l)

    return filename_output_l, filename_output_r



def fasta_patch(filename_fasta, filename_vcf, strain, filename_output, bgzip=False,
                num_processes=None, pass_only=False, quality=False, diploid=False):
    """
    Patch a Fasta file by replacing the bases where the SNPs are located in the VCF file.

    :param filename_fasta: name of the input Fasta file
    :type filename_fasta: string
    :param filename_vcf: name of the VCF file
    :type filename_vcf: string
    :param strain: name of strain to use in VCF file
    :type strain: string
    :param filename_output: name of the output Fasta file
    :type filename_output: string
    :param bgzip: compress file in BGZIP format
    :type bgzip: boolean
    :param num_processes: the number of processes to spawn
    :type num_processes: int
    :param pass_only: Only process those VCF records with a 'PASS'
    :type pass_only: boolean
    :param quality: filter on quality, FI=PASS
    :type quality: boolean
    :param diploid: don't ignore hets and create 2 files
    :type diploid: boolean
    :return: Nothing
    """
    start = time.time()

    filename_fasta = g2g_fu.check_file(filename_fasta)
    filename_vcf = g2g_fu.check_file(filename_vcf)

    LOG.info("INPUT FASTA FILE: {0}".format(filename_fasta))
    LOG.info("VCF FILE: {0}".format(filename_vcf))
    LOG.info("STRAIN: {0}".format(strain))
    LOG.info("PASS FILTER ON: {0}".format(str(pass_only)))
    LOG.info("QUALITY FILTER ON: {0}".format(str(quality)))
    LOG.info("DIPLOID: {0}".format(str(diploid)))

    if not strain:
        raise G2GValueError("No strain was specified.")

    filename_output_l, filename_output_r = prepare_fasta_patch(filename_fasta, filename_output, bgzip, diploid)

    if not num_processes:
        num_processes = multiprocessing.cpu_count()
    else:
        if num_processes <= 0:
            num_processes = 1

    LOG.info("NUMBER OF PROCESSES: {0}".format(num_processes))
    if bgzip:
        if diploid:
            LOG.info("OUTPUT FASTA FILES: {0}.gz".format(filename_output_l))
            LOG.info("                    {0}.gz".format(filename_output_r))
        else:
            LOG.info("OUTPUT FASTA FILE: {0}.gz".format(filename_output_l))
    else:
        if diploid:
            LOG.info("OUTPUT FASTA FILES: {0}".format(filename_output_l))
            LOG.info("                    {0}".format(filename_output_r))
        else:
            LOG.info("OUTPUT FASTA FILE: {0}".format(filename_output_l))

    LOG.info("Patching...")

    try:
        patch(filename_fasta, filename_vcf, strain, filename_output_l, filename_output_r,
              num_processes, pass_only, quality, diploid)

        LOG.info("Patching complete")

        # remove the fai
        LOG.debug("removing the FAI index for {0}".format(g2g_fu.delete_index_files(filename_output_l)))
        g2g_fu.delete_index_files(filename_output_l)

        # move temp to final destination
        if bgzip:
            LOG.info("Compressing and indexing...")
            g2g_fu.bgzip_index(filename_output_l, "{0}.gz".format(filename_output_l), 'fa')
            if diploid:
                g2g_fu.bgzip_index(filename_output_r, "{0}.gz".format(filename_output_r), 'fa')

        LOG.info("Execution complete: {0}".format(format_time(start, time.time())))
    except Exception, e:
        LOG.debug(e)
        raise G2GError("")





# -*- coding: utf-8 -*-

#
# Collection of functions related to chain files
#
# http://genome.ucsc.edu/goldenpath/help/chain.html
#


from collections import namedtuple
import os
import sys

from bx.intervals.intersection import Interval, IntervalTree

from .exceptions import G2GChainFileError
from .g2g_utils import configure_logging, get_logger, intersect_regions
from .g2g_fileutils import open_resource

LOG = get_logger()

# named tuple rather than class
IntervalMapping = namedtuple('IntervalMapping', ['from_chr', 'from_start', 'from_end', 'from_seq',
                                                 'to_chr', 'to_start', 'to_end', 'to_seq', 'same_bases', 'vcf_pos'])

CHAIN_STRING = "chain 1000 {from_chr} {from_length} + {from_start} {from_end} " + \
               "{to_chr} {to_length} + {to_start} {to_end} {id}"

CHAIN_HEADER_CHAIN = 0
CHAIN_HEADER_SCORE = 1
CHAIN_HEADER_TNAME = 2
CHAIN_HEADER_TSIZE = 3
CHAIN_HEADER_TSTRAND = 4
CHAIN_HEADER_TSTART = 5
CHAIN_HEADER_TEND = 6
CHAIN_HEADER_QNAME = 7
CHAIN_HEADER_QSIZE = 8
CHAIN_HEADER_QSTRAND = 9
CHAIN_HEADER_QSTART = 10
CHAIN_HEADER_QEND = 11
CHAIN_HEADER_ID = 12

class ChainEntry(object):
    def __init__(self):
        self.lines = []

    def __str__(self):
        return "\n".join(map(str, self.lines))


class ChainFile:
    """
    Encapsulate a chain file
    """
    def __init__(self, file_name, seq_ids=None, reverse=False):
        """
        Initialize a `.chain.ChainFile` object from a file.

        :param file_name: the name of the chain file
        :type file_name: string
        :param seq_ids: the ids of the sequences (chromosomes in most cases) to keep
        :type seq_ids: list
        :param reverse: `True` to parse in reverse
        :type reverse: boolean
        :return: and initialized `.chain.ChainFile` object
        """
        if not file_name:
            raise G2GChainFileError("Chain File must have a name")

        self.file_name = file_name
        self.dir, self.name = os.path.split(self.file_name)
        self.mapping_tree = None
        self.valid = False
        self.chrom_size_from = {}
        self.chrom_size_to = {}
        self.version = None
        self.headers = []
        self.seq_ids = seq_ids
        self.is_reversed = reverse

        self.parse(self.is_reversed)

    def get_seqids(self):
        if self.mapping_tree:
            return self.mapping_tree.keys()
        return None

    def parse(self, reverse):
        """
        Parse a chain file into mapping structure.

        The following documentation should explain a chain file format.::

            Header Line

            chain score tName tSize tStrand tStart tEnd qName qSize qStrand qStart qEnd id

            The initial header line starts with the keyword chain, followed by 11 required attribute values,
            and ending with a blank line. The attributes include:

            score -- chain score
            tName -- chromosome (reference sequence)
            tSize -- chromosome size (reference sequence)
            tStrand -- strand (reference sequence)
            tStart -- alignment start position (reference sequence)
            tEnd -- alignment end position (reference sequence)
            qName -- chromosome (query sequence)
            qSize -- chromosome size (query sequence)
            qStrand -- strand (query sequence)
            qStart -- alignment start position (query sequence)
            qEnd -- alignment end position (query sequence)
            id -- chain ID

            The alignment start and end positions are represented as zero-based half-open intervals. For example, the
            first 100 bases of a sequence would be represented with start position = 0 and end position = 100, and the
            next 100 bases would be represented as start position = 100 and end position = 200. When the strand value
            is "-", position coordinates are listed in terms of the reverse-complemented sequence.

            Alignment Data Lines

            Alignment data lines contain three required attribute values:

            size dt dq
            size -- the size of the ungapped alignment
            dt -- the difference between the end of this block and the beginning of the next block (reference sequence)
            dq -- the difference between the end of this block and the beginning of the next block (query sequence)

            NOTE: The last line of the alignment section contains only one number: the ungapped alignment size of
            the last block.

            Example:

            chain 4900 chrY 58368225 + 25985406 25985566 chr5 151006098 - 43549808 43549970 2
              16      0       2
              60      4       0
              10      0       4
              70

        :param file_name: the file to open and parse
        :param reverse: True to reverse
        :return: a ChainFile object
        """
        mapping_tree = {}
        chrom_size_from = {}
        chrom_size_to = {}
        chain_id_line = None
        line_no = 0
        process_seqid = False

        self.headers = []

        _chain_header_chain = CHAIN_HEADER_CHAIN
        _chain_header_score = CHAIN_HEADER_SCORE
        _chain_header_tname = CHAIN_HEADER_TNAME if not reverse else CHAIN_HEADER_QNAME
        _chain_header_tsize = CHAIN_HEADER_TSIZE if not reverse else CHAIN_HEADER_QSIZE
        _chain_header_tstrand = CHAIN_HEADER_TSTRAND if not reverse else CHAIN_HEADER_QSTRAND
        _chain_header_tstart = CHAIN_HEADER_TSTART if not reverse else CHAIN_HEADER_QSTART
        _chain_header_tend = CHAIN_HEADER_TEND if not reverse else CHAIN_HEADER_QEND
        _chain_header_qname = CHAIN_HEADER_QNAME if not reverse else CHAIN_HEADER_TNAME
        _chain_header_qsize = CHAIN_HEADER_QSIZE if not reverse else CHAIN_HEADER_TSIZE
        _chain_header_qstrand = CHAIN_HEADER_QSTRAND if not reverse else CHAIN_HEADER_TSTRAND
        _chain_header_qstart = CHAIN_HEADER_QSTART if not reverse else CHAIN_HEADER_TSTART
        _chain_header_qend = CHAIN_HEADER_QEND if not reverse else CHAIN_HEADER_TEND
        _chain_header_id = CHAIN_HEADER_ID

        _size = 0
        _dt = 1 if not reverse else 2
        _dq = 2 if not reverse else 1
        _same_bases = 3
        _dt_bases = 4 if not reverse else 5
        _dq_bases = 5 if not reverse else 4
        _vcf_pos = 6

        for line in open_resource(self.file_name):
            line = line.strip()
            line_no += 1

            if not line:
                continue

            if line.startswith(('#', ' ')):
                LOG.debug("Header: {0}".format(line))
                elems = line.split()
                if elems[0] == '#' and len(elems) > 1:
                    header_values = elems[1].strip().split('=')
                    if len(header_values) <= 1:
                        continue
                    self.headers.append((header_values[0].strip(), header_values[1].strip()))
                    if 'VERSION' == header_values[0].strip():
                        try:
                            self.version = int(header_values[1].strip())
                        except ValueError, e:
                            raise G2GChainFileError("Invalid chain file version '{0}'".format(self.version))
                continue

            fields = line.split()

            if fields[_chain_header_chain] == 'chain' and len(fields) in [12, 13]:
                # chain line
                LOG.debug(line)

                chain_id_line = line
                score = fields[_chain_header_score]

                # source
                t_name = fields[_chain_header_tname]

                if self.seq_ids:
                    if t_name in self.seq_ids:
                        LOG.debug("FOUND")
                        process_seqid = True
                    else:
                        process_seqid = False
                else:
                    process_seqid = True

                t_size = int(fields[_chain_header_tsize])
                t_strand = fields[_chain_header_tstrand]

                if t_strand != '+':
                    raise G2GChainFileError("tStrand in chain file must be '+'. Line #{0}: {1}".format(line_no, line))

                t_start = int(fields[_chain_header_tstart])
                t_end = int(fields[_chain_header_tend])

                #target
                q_name = fields[_chain_header_qname]
                q_size = int(fields[_chain_header_qsize])
                q_strand = fields[_chain_header_qstrand]

                if q_strand not in ['+', '-']:
                    raise G2GChainFileError("qStrand in chain file must be '+' or '-'. Line #{0}: {1}".format(line_no, line))

                q_start = int(fields[_chain_header_qstart])
                q_end = int(fields[_chain_header_qend])

                chrom_size_from[t_name] = t_size
                chrom_size_to[q_name] = q_size

                id = None if len(fields) == 12 else fields[12]

                if t_name not in mapping_tree:
                    mapping_tree[t_name] = IntervalTree()

                t_from, q_from = t_start, q_start

            elif fields[0] != 'chain' and (len(fields) == 3 or len(fields) == 7):
                # normal data line

                if not process_seqid:
                    continue

                if not self.version:
                    if len(fields) == 3:
                        self.version = 1
                    elif len(fields) == 7:
                        self.version = 2

                size, dt, dq = int(fields[_size]), int(fields[_dt]), int(fields[_dq])

                if len(fields) == 3:
                    if q_strand == '+':
                        mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_from, q_from+size, q_strand, None, None, None, None)))
                    elif q_strand == '-':
                        mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_size - (q_from+size), q_size - q_from, q_strand, None, None, None, None)))
                else:
                    if q_strand == '+':
                        mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_from, q_from+size, q_strand, fields[_dt_bases], fields[_dq_bases], fields[_same_bases], fields[_vcf_pos])))
                    elif q_strand == '-':
                        mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_size - (q_from+size), q_size - q_from, q_strand, fields[_dt_bases], fields[_dq_bases], fields[_same_bases], fields[_vcf_pos])))

                t_from += size + dt
                q_from += size + dq
            elif fields[0] != 'chain' and len(fields) == 1:
                # ending chain line
                if not process_seqid:
                    continue

                size = int(fields[0])

                if q_strand == '+':
                    mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_from, q_from+size, q_strand, None, None, None, None)))
                elif q_strand == '-':
                    mapping_tree[t_name].insert_interval(Interval(t_from, t_from+size, (q_name, q_size - (q_from+size), q_size - q_from, q_strand, None, None, None, None)))

                if (t_from + size) != t_end or (q_from + size) != q_end:
                    LOG.debug("{0} + {1} != {2}".format(t_from, size, t_end))
                    LOG.debug("or")
                    LOG.debug("{0} + {1} != {2}".format(q_from, size, q_end))
                    raise G2GChainFileError("Alignment blocks do not match specified block sizes. {0}".format(chain_id_line))
            else:
                raise G2GChainFileError("Unexpected line in chain file. Line #{0:,}: {1}".format(line_no, line))

        LOG.debug("PROCESSED {0:,} lines".format(line_no))
        self.valid = True
        self.mapping_tree = mapping_tree
        self.chrom_size_from = chrom_size_from
        self.chrom_size_to = chrom_size_to

    def find_mappings(self, chromosome, start, end):
        """
        Find mapping from source to target

        :param chromosome: chromosome
        :param start: start position
        :param end: end position
        :param strand: strand
        :return: list of IntervalMappings
        """
        mappings = []
        #LOG.debug("Finding mappings for {0}:{1}-{2}".format(chromosome, start, end))

        if chromosome not in self.mapping_tree:
            return None

        all_intervals = self.mapping_tree[chromosome].find(start, end)

        if len(all_intervals) == 0:
            #LOG.debug("No intervals found")
            return None
        else:
            for interval in all_intervals:
                #LOG.debug("Interval {0}={1}".format(len(mappings), interval))
                chromosome, real_start, real_end = intersect_regions(chromosome, start, end,
                                                                     chromosome, interval.start, interval.end)
                offset = abs(real_start - interval.start)
                size = abs(real_end - real_start)
                i_start = interval.value[1] + offset
                mappings.append(IntervalMapping(chromosome, real_start, real_end, interval.value[4],
                                                interval.value[0], i_start, i_start + size, interval.value[5], interval.value[6], interval.value[7]))

                #LOG.debug("Mapping {0}={1}".format(len(mappings)-1, mappings[-1]))

        return mappings


def collapse_entries(entries):
    """
    Recursive function to collapse chain entries.

    Example:

        [10, 20, 0]
        [2, 5, 0]
        [10]

        [5]

        [4, 0, 2]
        [10]

    would collapse to

        [10, 20, 0]
        [2, 5, 0]
        [19, 0, 2]
        [10]

    :param entries: array of chain file entries
    :return: the collapsed entries
    """
    if len(entries) == 1:
        LOG.debug('base case')
        return entries

    # combine first 2 entries
    entry1 = entries[0]
    entry2 = entries[1]

    last_line = entry1.lines[-1]
    first_line = entry2.lines[0]

    LOG.debug("last_line={0}".format(last_line))
    LOG.debug("first_line={0}".format(first_line))

    new_line = []
    lines = []

    if len(last_line) == 1 and len(first_line) == 1:
        new_line.append(last_line[0] + first_line[0])

        if len(entry1.lines) > 1:
            lines.extend(entry1.lines[:-1])

        lines.append(new_line)

    elif len(last_line) == 1 and len(first_line) != 1:

        new_line.append(last_line[0] + first_line[0])
        new_line.append(first_line[1])
        new_line.append(first_line[2])

        if len(entry1.lines) > 1:
            lines.extend(entry1.lines[:-1])
        lines.append(new_line)
        lines.extend(entry2.lines[1:])
    elif len(last_line) != 1 and len(first_line) == 1:
        LOG.error("ERROR")
    else:
        LOG.error("ERROR")

    new_entry = ChainEntry()
    new_entry.lines = lines

    new_entries = []
    new_entries.append(new_entry)
    new_entries.extend(entries[2:])

    return collapse_entries(new_entries)


class ChainIter(object):
    """
    Simple way to iterate chain file
    """
    def __init__(self, file_name, reverse=False):
        if not file_name:
            raise G2GChainFileError("A filename must be supplied")

        self.file_name = file_name
        self.reverse = reverse
        self.current_chain_header = None
        self.current_line = None
        self.current_record = None
        self.line_no = 0
        self.reader = open_resource(file_name)

    def reset(self):
        if self.reader:
            try:
                self.reader.close()
            except:
                pass

        self.current_chain_header = None
        self.current_line = None
        self.current_record = None
        self.line_no = 0
        self.reader = open_resource(self.file_name)


    def __iter__(self):
        return self

    def next(self):
        self.current_line = self.reader.next()
        self.line_no += 1

        while self.current_line.startswith("#"):
            self.line_no += 1
            self.current_line = self.reader.next()

        while len(self.current_line.strip()) == 0:
            self.line_no += 1
            self.current_line = self.reader.next()

        self.current_record = self.parse_line(self.current_line)

        return self.current_record

    def parse_line(self, line):
        if not line:
            raise G2GChainFileError("Empty line")

        fields = line.split()

        _chain_header_chain = CHAIN_HEADER_CHAIN
        _chain_header_score = CHAIN_HEADER_SCORE
        _chain_header_tname = CHAIN_HEADER_TNAME if not self.reverse else CHAIN_HEADER_QNAME
        _chain_header_tsize = CHAIN_HEADER_TSIZE if not self.reverse else CHAIN_HEADER_QSIZE
        _chain_header_tstrand = CHAIN_HEADER_TSTRAND if not self.reverse else CHAIN_HEADER_QSTRAND
        _chain_header_tstart = CHAIN_HEADER_TSTART if not self.reverse else CHAIN_HEADER_QSTART
        _chain_header_tend = CHAIN_HEADER_TEND if not self.reverse else CHAIN_HEADER_QEND
        _chain_header_qname = CHAIN_HEADER_QNAME if not self.reverse else CHAIN_HEADER_TNAME
        _chain_header_qsize = CHAIN_HEADER_QSIZE if not self.reverse else CHAIN_HEADER_TSIZE
        _chain_header_qstrand = CHAIN_HEADER_QSTRAND if not self.reverse else CHAIN_HEADER_TSTRAND
        _chain_header_qstart = CHAIN_HEADER_QSTART if not self.reverse else CHAIN_HEADER_TSTART
        _chain_header_qend = CHAIN_HEADER_QEND if not self.reverse else CHAIN_HEADER_TEND
        _chain_header_id = CHAIN_HEADER_ID

        _size = 0
        _dt = 1 if not self.reverse else 2
        _dq = 2 if not self.reverse else 1
        _same_bases = 3
        _dt_bases = 4 if not self.reverse else 5
        _dq_bases = 5 if not self.reverse else 4
        _vcf_pos = 6

        if fields[CHAIN_HEADER_CHAIN] == 'chain' and len(fields) in [12, 13]:
            # chain line
            LOG.debug(line)

            chain_id_line = line
            score = fields[_chain_header_score]

            # source
            t_name = fields[_chain_header_tname]
            t_size = int(fields[_chain_header_tsize])
            t_strand = fields[_chain_header_tstrand]

            if t_strand != '+':
                raise G2GChainFileError("tStrand in chain file must be '+'. Line #{0}: {1}".format(self.line_no, line))

            t_start = int(fields[_chain_header_tstart])
            t_end = int(fields[_chain_header_tend])

            #target
            q_name = fields[_chain_header_qname]
            q_size = int(fields[_chain_header_qsize])
            q_strand = fields[_chain_header_qstrand]

            if q_strand not in ['+', '-']:
                raise G2GChainFileError("qStrand in chain file must be '+' or '-'. Line #{0}: {1}".format(self.line_no, line))

            q_start = int(fields[_chain_header_qstart])
            q_end = int(fields[_chain_header_qend])

            id = None if len(fields) == 12 else fields[12]

            self.current_chain_header = [score, t_name, t_size, t_strand, t_start, t_end, q_name, q_size, q_strand, q_start, q_end, id]

            return self.current_chain_header
        elif fields[0] != 'chain' and (len(fields) == 3 or len(fields) == 7):
            # normal data line
            return [int(fields[_size]), int(fields[_dt]), int(fields[_dq]), fields[_same_bases], fields[_dt_bases], fields[_dq_bases], fields[_vcf_pos]]
        elif fields[0] != 'chain' and len(fields) == 1:
            return [int(fields[0])]



def abc(a):
    print a

if __name__ == '__main__':
    """
    #print 'a'
    #c = ChainFile(sys.argv[1], 2, '11')
    #print 'hi'
    import struct
    fd = open('akjahs', "w")
    for c in xrange(20):
        print str(c)
        for i in xrange(2000000):
            a = (str(c), i, i*2, str(c), i+1, i*2+1, "ACGGT", "AAD")
            struct.pack(a)
    fd.close()
    """
    configure_logging(5)
    for r in [False, True]:
        c = ChainFile(sys.argv[1], 2, reverse=r)
        c.mapping_tree['1'].traverse(abc)
        #print ""
        #print '1', 3000015, 3000099
        #print c.find_mappings('1', 3000015, 3000099)
        #print ""
        #print '1', 3000015, 3000099
        #print c.find_mappings('1', 3000015, 3000099)
        #print ""
        #print '1', 3000099, 3000015
        #print c.find_mappings('1', 3000099, 3000015)
        #print ""
        #print '1', 3000099, 3000015
        #print c.find_mappings('1', 3000099, 3000015)
        #print ""
        print 'done'





# -*- coding: utf-8 -*-

from .exceptions import G2GError, G2GBAMError, G2GFastaError, G2GFetchError, G2GCigarFormatError, G2GChainFileError,\
    G2GBedError, G2GVCFError, G2GLocationError, G2GValueError, G2GGTFError
from .g2g import file_convert, offset2chain, gtf2chain, db2chain, fasta_transform
from .g2g_commands import command_convert, command_offset2chain, command_gtf2chain, command_gtf2db, command_db2chain, \
    command_vcf2chain, command_fasta_extract, command_fasta_patch, command_fasta_transform, command_parse_location
from .g2g_utils import parse_location, complement_sequence, reverse_sequence, reverse_complement_sequence
from .gtf_db import gtf2db
from .chain import ChainFile
from .fasta import extract as fasta_extract
from .fasta_patch import fasta_patch
from .vcf2chain import vcf2chain
from .bins import bins


__version__ = '0.1.13'
__author__ = 'Matthew Vincent, The Jackson Laboratory'
__email__ = 'matt.vincent@jax.org'


__all__ = ['G2GError', 'G2GBAMError', 'G2GFastaError', 'G2GFetchError', 'G2GCigarFormatError', 'G2GChainFileError',
           'G2GBedError', 'G2GVCFError', 'G2GLocationError', 'G2GValueError', 'G2GGTFError',
           'file_convert', 'offset2chain', 'gtf2chain', 'gtf2db', 'db2chain', 'fasta_transform',
           'command_convert', 'command_offset2chain', 'command_gtf2chain', 'command_gtf2db',
           'command_db2chain', 'command_vcf2chain', 'command_fasta_extract', 'command_fasta_transform',
           'command_fasta_patch', 'command_parse_location',
           'parse_location', 'complement_sequence', 'reverse_sequence', 'reverse_complement_sequence',
           'gtf2db',
           'ChainFile',
           'fasta_extract',
           'fasta_patch',
           'vcf2chain'
           'bins']




# -*- coding: utf-8 -*-

#
# Collection of functions related to FASTA files
#

import difflib
import sys
import time
import os

from collections import OrderedDict


try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from .exceptions import G2GFastaError, G2GLocationError
from .g2g_utils import Location, format_time, get_logger, complement_sequence, reverse_sequence, reverse_complement_sequence, wrap_sequence
from .g2g_fileutils import check_file

import pysam

LOG = get_logger()


class FAIEntry(object):
    def __init__(self, idx=None, length=0, offset=-1, line_length=None, line_length_bytes=None):
        """
        Initialize a `.fasta.FAIEntry` object.

        :param idx: index of fasta sequence
        :type idx: string
        :param length: length of fasta sequence
        :type length: int
        :param offset: offset in file
        :type offset: int
        :param line_length: line length
        :type line_length: int
        :param line_length_bytes: line length + newline
        :type line_length_bytes: int
        :return: `.fasta.FAIEntry`
        """
        self.idx = idx
        self.length = length
        self.offset = offset
        self.line_length = line_length
        self.line_length_bytes = line_length_bytes

    def __str__(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}".format(self.idx, self.length, self.offset, self.line_length, self.line_length_bytes)


class FAI(object):
    def __init__(self, fasta_filename):
        """
        Initialize a `.fasta.FAI` file, an index into the Fasta file.

        :param file_name: the name of the Fasta file
        :type file_name: string
        :return: an initialized `.fasta.FAI` object
        """
        self.fasta_file = fasta_filename
        self.index_file = "{0}.fai".format(fasta_filename)
        self.records = OrderedDict()

        if os.path.exists(self.index_file):
            self.read()
        else:
            raise G2GFastaError("Unable to find fasta index file")

    def __getitem__(self, entry):
        try:
            if isinstance(entry, int):
                entry = tuple(self.records.keys())[entry]

            return self.records[entry]
        except KeyError:
            raise G2GFastaError("{0} not in {1}.".format(entry, self.index_file))

    def __iter__(self):
        for keys in self.records:
            yield keys

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__, self.index_file)

    def read(self):
        with open(self.index_file) as index:
            for line in index:
                line = line.strip()
                idx, length, offset, line_length, line_length_bytes = line.split('\t')
                if idx in self.records:
                    raise ValueError("A duplicate entry was found: {0}".format(idx))
                else:
                    self.records[idx] = FAIEntry(idx, int(length), int(offset), int(line_length), int(line_length_bytes))


def get(fasta, location):
    """
    0-based,exclusive extraction from a Fasta file.

    ACGTACGTACGTACGT

    chr1:4-9, would yield

    ----ACGTA-------
    0123456789012345

    ACGTA

    chr1:1, would yield

    -C--------------
    0123456789012345

    C

    chr1:1-2, would yield

    -C--------------
    0123456789012345

    C (Not CG, because it is exclusive).  If you wanted CG, you would request chr1:1-3.
    """

    try:
        if location.end and location.start == location.end:
            return fasta[location.seqid][location.start]
        elif not location.end:
            return fasta[location.seqid][location.start]
        else:
            return fasta[location.seqid][location.start:location.end]
    except G2GLocationError as e:
        LOG.info(e.msg.rstrip())
        raise e
    except G2GFastaError as e:
        LOG.info(e.msg.rstrip())
        raise e

    return None


def extract(fasta_input, locations, output=None, reverse=False, complement=False, raw=False):
    """

    :param fasta_input: the name of the Fasta file
    :type fasta_input: string
    :param locations: list of `.g2g_utils.Location` objects
    :type locations: list
    :param output: name of output file, None for sys.stdout
    :type output: string
    :param reverse: reverse the extracted sequence
    :type reverse: boolean
    :param complement: complement the extracted sequence
    :type complement: boolean
    :param raw: for just the sequence to be output, no fasta line
    :type raw: boolean
    """

    start_time = time.time()

    if isinstance(fasta_input, pysam.FastaFile):
        fasta = fasta_input
    else:
        fasta_input = check_file(fasta_input)
        fasta = pysam.FastaFile(fasta_input)

    LOG.debug("FASTA FILE: {0}".format(fasta.filename))

    if output:
        output = check_file(output, 'w')
        fasta_out = open(output, "w")
    else:
        fasta_out = sys.stdout

    try:
        if not isinstance(locations, list):
            locations = [locations]

        for location in locations:
            LOG.debug('LOCATION: {0}'.format(location))
            if not isinstance(location, Location):
                raise G2GLocationError("Found: {0}, which is not a Location")

            if location.seqid not in fasta.references:
                continue

            sequence = fasta.fetch(location.seqid, location.start, location.end)

            if location.strand == '+':
                if reverse and not complement:
                    sequence = reverse_sequence(sequence)
                elif not reverse and complement:
                    sequence = complement_sequence(sequence)
                elif reverse and complement:
                    sequence = reverse_complement_sequence(sequence)
            else:
                sequence = reverse_complement_sequence(sequence)

            if raw:
                fasta_out.write(sequence)
                fasta_out.write('\n')
            else:
                # internally storing 0 base, outputting 1 base
                start = location.start + 1
                end = location.end

                fasta_id = ">{0} {1}-{2}\n".format(location.seqid, start, end)
                LOG.debug(fasta_id)
                LOG.debug("{0}...{1}".format(sequence[:10], sequence[-10:]))

                if location.name:
                    fasta_id = ">{0} {1}:{2}-{3}\n".format(location.name, location.seqid, start, end)

                fasta_out.write(fasta_id)

                for line in wrap_sequence(sequence):
                    fasta_out.write(line.strip())
                    fasta_out.write('\n')

    except G2GLocationError as e:
        LOG.info(e.msg.rstrip())
        raise e
    except G2GFastaError as e:
        LOG.info(e.msg.rstrip())
        raise e

    fasta_out.close()

    LOG.info("Execution complete: {0}".format(format_time(start_time, time.time())))


def extract_id(fasta_input, identifier, output=None, reverse=False, complement=False, raw=False):
    """
    """
    LOG.info("Execution start")

    start_time = time.time()

    if isinstance(fasta_input, pysam.FastaFile):
        fasta = fasta_input
    else:
        fasta_input = check_file(fasta_input)
        fasta = pysam.FastaFile(fasta_input)

    LOG.debug("FASTA FILE: {0}".format(fasta.filename))

    if output:
        output = check_file(output, 'w')
        fasta_out = open(output, "w")
    else:
        fasta_out = sys.stdout

    try:
        sequence = fasta[identifier]

        if reverse and not complement:
            sequence = reverse_sequence(sequence)
        elif not reverse and complement:
            sequence = complement_sequence(sequence)
        elif reverse and complement:
            sequence = reverse_complement_sequence(sequence)

        if raw:
            fasta_out.write(sequence)
        else:
            fasta_id = ">{0}\n".format(identifier)
            fasta_out.write(fasta_id)

            for line in wrap_sequence(sequence):
                fasta_out.write(line.strip())
                fasta_out.write('\n')

    except G2GLocationError as e:
        LOG.info(e.msg.rstrip())
        raise e
    except G2GFastaError as e:
        LOG.info(e.msg.rstrip())
        raise e

    fasta_out.close()

    LOG.info("Execution complete: {0}".format(format_time(start_time, time.time())))


def diff(sequence1, sequence2):
    """

    cases=[('afrykanerskojęzyczny', 'afrykanerskojęzycznym'),
           ('afrykanerskojęzyczni', 'nieafrykanerskojęzyczni'),
           ('afrykanerskojęzycznym', 'afrykanerskojęzyczny'),
           ('nieafrykanerskojęzyczni', 'afrykanerskojęzyczni'),
           ('nieafrynerskojęzyczni', 'afrykanerskojzyczni'),
           ('abcdefg','xac')]

    for a,b in cases:
        print('{} => {}'.format(a,b))
        for i,s in enumerate(difflib.ndiff(a, b)):
            if s[0]==' ': continue
            elif s[0]=='-':
                print(u'Delete "{}" from position {}'.format(s[-1],i))
            elif s[0]=='+':
                print(u'Add "{}" to position {}'.format(s[-1],i))
        print()
    """
    for i, s in enumerate(difflib.ndiff(sequence1, sequence2)):
        print i, s

        if s[0]==' ':
            continue
        elif s[0] == '-':
            print(u"Delete '{}' from position {}".format(s[-1], i))
        elif s[0] == '+':
            print(u"Add '{}' to position {}".format(s[-1], i))
    print()


def compare_files(file_1, file_2, enid):
    f1 = pysam.FastaFile(file_1)
    f2 = pysam.FastaFile(file_2)

    fa_equal={}
    fa_diff={}


    c = 0

    if enid:
        pass



    else:
        for ensembl_id in f1:
            seq_1 = f1[ensembl_id]
            seq_2 = f2[ensembl_id]
            if len(str(seq_1)) == len(str(seq_2)):
            #if str(seq_1) == str(seq_2):
                fa_equal[ensembl_id] = {'seq_1':str(seq_1), 'seq_2':str(seq_2)}
            else:
                fa_diff[ensembl_id] = {'seq_1':str(seq_1), 'seq_2':str(seq_2)}
                if abs(len(str(seq_1)) - len(str(seq_2))) > 30:
                    print ensembl_id

    print "# EQUAL: {0}".format(len(fa_equal))
    print "# DIFF: {0}".format(len(fa_diff))





if __name__ == '__main__':
    #diff(sys.argv[1], sys.argv[2])
    compare_files(sys.argv[1], sys.argv[2], sys.argv[3] if len(sys.argv) == 4 else None)
    #create_compare_fasta_file(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])




# -*- coding: utf-8 -*-

#
# A way to create a Chain file from a VCF file quickly
#

from collections import OrderedDict
from collections import Counter
import multiprocessing
import os
import sys
import time

import pysam
from pysam import TabixFile
from pysam import FastaFile

from .g2g_utils import format_time, get_logger
import g2g_fileutils as g2g_fu
from . import G2GValueError, G2GVCFError, G2GError, G2GChainFileError
from .vcf import parse_gt_new
from .chain import CHAIN_STRING
import traceback

LOG = get_logger()


class VCFtoChainInfo(object):
    def __init__(self):
        self.chromosome = None
        self.chromosome_length = 0
        self.number_vcf_lines = 0
        self.last_fragment_size = 0
        self.prev_next_ref_pos = 1
        self.chain_entries = []
        self.sums = [0, 0, 0]
        self.stats = {}
        self.end_length = 0
        self.output_file = None


class KeyboardInterruptError(Exception):
    pass


def _show_error():
    """
    show system errors
    """
    et, ev, tb = sys.exc_info()

    print "Error Type: %s" % et
    print "Error Value: %s" % ev
    while tb :
        co = tb.tb_frame.f_code
        filename = str(co.co_filename)
        line_no = str(traceback.tb_lineno(tb))
        print '    %s:%s' % (filename, line_no)
        tb = tb.tb_next


def update_stats(stats, reason):
    if not stats:
        stats = OrderedDict()

    if reason in stats:
        stats[reason] += 1
    else:
        stats[reason] = 1

    return stats


def process_piece(filename_vcf, chrom, chrom_length, sample_index, chain_info, diploid, passed, quality, vcf_keep, vcf_discard_file):
    ret = {'chrom': chrom, 'stats': {}, 'chain_info': chain_info}

    stats = OrderedDict()
    stats['ACCEPTED'] = 0

    if vcf_keep:
        vcf_discard = open(vcf_discard_file, "w")

    line_no = 0

    try:
        LOG.info("Processing Chromosome {0}...".format(chrom))
        tb = pysam.TabixFile(filename_vcf)

        for vcf_rec in tb.fetch(chrom, parser=pysam.asVCF()):
            line_no += 1

            try:
                gt = parse_gt_new(vcf_rec, sample_index)
            except:
                LOG.info("Unable to parse record, improper VCF file?")
                continue

            LOG.debug('\n')
            LOG.debug(vcf_rec)
            LOG.debug(gt)
            LOG.debug(vcf_rec[sample_index])

            if passed and 'PASS' not in vcf_rec.FILTER:

                LOG.debug("TOSSED: FILTERED ON PASS")
                stats = update_stats(stats, 'FILTERED ON PASS')

                if vcf_keep:
                    vcf_discard.write(vcf_rec)
                    vcf_discard.write("\n")
                continue

            elif quality and gt.fi == '0':

                # FI : Whether a sample was a Pass(1) or fail (0) based on FILTER values

                LOG.debug("TOSSED: FILTERED ON QUALITY")
                stats = update_stats(stats, 'FILTERED ON QUALITY')

                if vcf_keep:
                    vcf_discard.write(vcf_rec)
                    vcf_discard.write("\n")
                continue

            elif gt.left is None and gt.right is None:

                LOG.debug("TOSSED: NOT RELEVANT")
                stats = update_stats(stats, 'NOT RELEVANT')

                if vcf_keep:
                    vcf_discard.write(vcf_rec)
                    vcf_discard.write("\n")
                continue

            elif not diploid and gt.left != gt.right:
                # haploid or hexaploid
                # gt must be equal

                LOG.debug("TOSSED: HETEROZYGOUS")
                stats = update_stats(stats, 'HETEROZYGOUS')

                if vcf_keep:
                    vcf_discard.write(vcf_rec)
                    vcf_discard.write("\n")
                continue

            # START L AND R, ONLY R IF DIPLOID

            for ci, lr in chain_info.iteritems():
                if ci == 'left':
                    alt_seq = str(gt.left)
                else:
                    alt_seq = str(gt.right)

                if gt.ref == alt_seq:

                    LOG.debug("TOSSED, SAME AS REF")
                    lr.stats = update_stats(lr.stats, 'SAME AS REF')

                    if vcf_keep:
                        vcf_discard.write(vcf_rec)
                        vcf_discard.write("\n")
                    continue

                orig_alt_seq = alt_seq

                LOG.debug("SAMPLE: {0}".format(vcf_rec[sample_index]))
                LOG.debug("REF='{0}', ALT_L='{1}', ALT_R='{2}'. POS={3}".format(gt.ref, gt.left, gt.right, vcf_rec.pos))

                position = vcf_rec.pos + 1

                ref_seq = str(gt.ref)
                len_ref = len(ref_seq)
                len_alt = len(alt_seq)

                base_changes = len_ref - len_alt
                base_pos_diff = 0

                if position < lr.prev_next_ref_pos:
                    LOG.debug("TOSSED: CONFLICTING VCF ENTRIES: {0}".format(vcf_rec))

                    lr.stats = update_stats(lr.stats, 'CONFLICTING VCF ENTRIES')

                    if vcf_keep:
                        vcf_discard.write(vcf_rec)
                        vcf_discard.write("\n")

                    continue

                # find the position where the first base change is
                for n in xrange(min(len_ref, len_alt)):
                    if ref_seq[n] != alt_seq[n]:
                        base_pos_diff = n
                        break

                # if it is 0, take the minimum length
                if base_pos_diff == 0:
                    base_pos_diff = min(len_ref, len_alt)

                # add the base position difference
                position += base_pos_diff

                # recalculate the strings
                shared_bases = ref_seq[:base_pos_diff]
                ref_seq = ref_seq[base_pos_diff:]
                alt_seq = alt_seq[base_pos_diff:]

                dt = len(ref_seq)
                dq = len(alt_seq)

                next_ref_pos = position + len(ref_seq)
                fragment_size = position - lr.prev_next_ref_pos

                LOG.debug('           gt.ref: {0}'.format(gt.ref))
                LOG.debug('          ref_seq: {0}'.format(ref_seq))
                LOG.debug('               dt: {0}'.format(dt))
                LOG.debug('           gt.alt: {0}'.format(orig_alt_seq))
                LOG.debug('          alt_seq: {0}'.format(alt_seq))
                LOG.debug('               dq: {0}'.format(dq))
                LOG.debug('         position: {0}'.format(position))
                LOG.debug('prev_next_ref_pos: {0}'.format(lr.prev_next_ref_pos))
                LOG.debug('     next_ref_pos: {0}'.format(next_ref_pos))
                LOG.debug('    fragment_size: {0}'.format(fragment_size))
                LOG.debug('     base_changes: {0}'.format(base_changes))
                LOG.debug('    base_pos_diff: {0}'.format(base_pos_diff))
                LOG.debug('     shared_bases: {0}'.format(shared_bases))

                # fix any 0 length
                if fragment_size < 0:
                    LOG.debug("TOSSED: CONFLICTING VCF ENTRIES: {0}".format(vcf_rec))

                    lr.stats = update_stats(lr.stats, 'CONFLICTING VCF ENTRIES')

                    if vcf_keep:
                        vcf_discard.write(vcf_rec)
                        vcf_discard.write("\n")

                    continue

                if fragment_size != 0:
                    ref_str = ref_seq if ref_seq else '.'
                    alt_str = alt_seq if alt_seq else '.'
                    lr.chain_entries.append([fragment_size, len(ref_seq), len(alt_seq), shared_bases, ref_str, alt_str, vcf_rec.pos+1])
                else:
                    #
                    # THIS SHOULD NOT HAPPEN
                    #
                    raise G2GChainFileError('Unable to create chain file due to conflicting VCF entries')

                lr.stats = update_stats(lr.stats, 'ACCEPTED')

                LOG.debug(lr.chain_entries[-1])

                last_position = position
                lr.prev_next_ref_pos = next_ref_pos
                lr.sums[0] += fragment_size
                lr.sums[1] += dt
                lr.sums[2] += dq
                prev_line = vcf_rec

                #lr.prev_chrom = vcf_rec.contig
            chain_info[ci] = lr

        for ci, lr in chain_info.iteritems():
            #LOG.debug("CHROMOSOME[{0}] LENGTH = {1}".format(lr.prev_chrom, chrom_length))

            lr.chromosome = chrom
            lr.chromosome_length = chrom_length
            lr.last_fragment_size = chrom_length - lr.sums[0] - lr.sums[1]
            lr.end_length = lr.sums[0] + lr.last_fragment_size + lr.sums[2]
            lr.number_vcf_lines = line_no

            chain_info[ci] = lr

        if vcf_keep:
            vcf_discard.close()

    except KeyboardInterrupt:
        raise KeyboardInterruptError()
    except Exception, e:
        pass

    ret['stats'] = stats
    ret['chain_info'] = chain_info

    return ret


def wrapper(args):
    """
    Simple wrapper, useful for debugging.

    :param args: the arguments to process_piece
    :return: the same as process_piece
    """
    return process_piece(*args)


def log_stats(stats):
    LOG.info("STATISTICS")
    for s, stat in stats.iteritems():
        LOG.info("{0:,}\t\t{1}".format(stat, s))


def vcf2chain(input_file, fasta_file, strain, output_file, vcf_keep=False, passed=False, quality=False, diploid=False):
    """

    :param input_file:
    :param fasta_file:
    :param strain:
    :param output_file:
    :param vcf_keep:
    :param passed:
    :param quality:
    :param diploid:
    :return:
    """
    start = time.time()

    input_file = g2g_fu.check_file(input_file)
    fasta_file = g2g_fu.check_file(fasta_file)

    if not strain:
        raise G2GValueError("No strain was specified.")

    output_file = g2g_fu.check_file(output_file, 'w')
    output_file_dir = os.path.dirname(output_file)

    LOG.info("VCF FILE: {0}".format(input_file))
    LOG.info("FASTA FILE: {0}".format(fasta_file))
    LOG.info("CHAIN FILE: {0}".format(output_file))

    vcf_discard_file = None

    if vcf_keep:
        vcf_discard_file = "{0}.errors.vcf".format(os.path.basename(input_file))
        vcf_discard_file = os.path.join(output_file_dir, vcf_discard_file)
        LOG.info("VCF DISCARD FILE: {0}".format(vcf_discard_file))

    LOG.info("STRAIN: {0}".format(strain))
    LOG.info("PASS FILTER ON: {0}".format(str(passed)))
    LOG.info("QUALITY FILTER ON: {0}".format(str(quality)))
    LOG.info("DIPLOID: {0}".format(str(diploid)))

    if not isinstance(fasta_file, FastaFile):
        fasta_file = FastaFile(fasta_file)

    tb = TabixFile(input_file)
    sample_index = None

    for h in tb.header:
        if h[:6] == '#CHROM':
            try:
                elems = h.split('\t')
                samples = elems[9:]
                samples = dict(zip(samples, (x for x in xrange(len(samples)))))
                sample_index = samples[strain]
            except KeyError, ke:
                raise G2GVCFError("Unknown strain '{0}', valid strains are: {1}".format(strain, ", ".join(samples)))

    LOG.info("STRAIN SAMPLE INDEX: {0}".format(sample_index))

    LOG.info("Parsing VCF file...")

    # not all chromosomes/seqid will be processed if not in vcf file
    processed_seqids = OrderedDict()

    for seqid in tb.contigs:
        processed_seqids[seqid] = False

    left = VCFtoChainInfo()
    right = VCFtoChainInfo()

    chain_info = {}

    if diploid:
        left.output_file = g2g_fu.prepend_before_extension(output_file, 'left')
        right.output_file = g2g_fu.prepend_before_extension(output_file, 'right')
        chain_info['left'] = left
        chain_info['right'] = right

        g2g_fu.delete_file(left.output_file)
        g2g_fu.delete_file(right.output_file)
    else:
        left.output_file = output_file
        chain_info['left'] = left

        g2g_fu.delete_file(left.output_file)

    try:
        all_chrom = [c for c in fasta_file.references]
        all_chrom_length = [n for n in fasta_file.lengths]
        all_vcffiles = [input_file] * len(all_chrom)
        all_sample_index = [sample_index] * len(all_chrom)
        all_chain_info = [chain_info] * len(all_chrom)
        all_diploid = [diploid] * len(all_chrom)
        all_passed = [passed] * len(all_chrom)
        all_quality = [quality] * len(all_chrom)
        all_vcf_keep = [vcf_keep] * len(all_chrom)
        all_vcf_discard_file = [vcf_discard_file] * len(all_chrom)

        args = zip(all_vcffiles, all_chrom, all_chrom_length, all_sample_index, all_chain_info, all_diploid, all_passed,
                   all_quality, all_vcf_keep, all_vcf_discard_file)

        pool = multiprocessing.Pool(8)
        #pool = multiprocessing.Pool(1)
        results = pool.map(wrapper, args)

        # parse results
        total = 0

        # show stats
        for c in results:
            LOG.info("Chromosome: {0}".format(c['chrom']))
            for ci, lr in c['chain_info'].iteritems():
                A = Counter(c['stats'])
                B = Counter(lr.stats)
                log_stats(A+B)
            total += c['chain_info']['left'].number_vcf_lines

        # create header of chain file
        for c in results:
            for ci, lr in c['chain_info'].iteritems():
                outfile = open(lr.output_file, 'a')
                write = outfile.write
                write("# CHAINFILE\n")
                write("# CREATED={0}\n".format(time.strftime("%m-%d-%Y %H:%M:%S")))
                write("# PROGRAM=vcf2chain\n")
                write("# VCF={0}\n".format(input_file))
                write("# FASTA={0}\n".format(fasta_file.filename))

            break

        # loop through the results and dump to file
        for c in results:
            for ci, lr in c['chain_info'].iteritems():
                outfile = open(lr.output_file, 'a')
                write = outfile.write

                if lr.number_vcf_lines > 0:
                    write(CHAIN_STRING.format(CHAIN_STRING,
                                from_chr=c['chrom'], from_length=lr.chromosome_length,
                                from_start=0, from_end=lr.chromosome_length,
                                to_chr=c['chrom'], to_length=lr.end_length,
                                to_start=0, to_end=lr.sums[0] + lr.last_fragment_size + lr.sums[2], id=c['chrom']))
                    write("\n")

                    for c in lr.chain_entries:
                        write("\t".join(map(str, c)))
                        write("\n")
                    write(str(lr.last_fragment_size))
                    write("\n\n")
                    outfile.close()
                else:
                    chrom_length = fasta_file.get_reference_length(c['chrom'])
                    write(CHAIN_STRING.format(CHAIN_STRING,
                        from_chr=c['chrom'], from_length=chrom_length,
                        from_start=0, from_end=chrom_length,
                        to_chr=c['chrom'], to_length=chrom_length,
                        to_start=0, to_end=chrom_length, id=c['chrom']))

                    write("\n")
                    write(str(chrom_length))
                    write("\n\n")

        LOG.info("complete")

        if vcf_keep:
            vcf_discard_file.close()

        LOG.info("Parsed {0:,} total lines".format(total))
        LOG.info("Chain file created")
        LOG.info("Execution complete: {0}".format(format_time(start, time.time())))
    except KeyboardInterrupt:
        pool.terminate()
        raise G2GError("Execution halted")
    except Exception, e:
        _show_error()
        raise G2GError("Execution halted")

